#!/usr/bin/env python  
# -*- coding: utf-8 -*-
#
#   convertMesh.py
#
#       メッシュの変換
#
#   18/02/03    新規作成
#

import geometryFunc as geo

ls = "\n"

#  deleteSameData
#    同じdataを削除する
def deleteSameData(dataList):
    dataList.sort()
    newData = [dataList[0]]
    for data in dataList:
        if data != newData[-1]:
            newData.append(data)
    return newData

#
#  nodeNoの並び順を変更する関数群
def order_3(el, n):
    element = [el[n[0]-1], el[n[1]-1], el[n[2]-1]]
    return element

def order_4(el, n):
    element = [el[n[0]-1], el[n[1]-1], el[n[2]-1], el[n[3]-1]]
    return element

def order_6(el):
    element = [el[n[0]-1], el[n[1]-1], el[n[2]-1], el[n[3]-1], el[n[4]-1],
               el[n[5]-1]]
    return element

def order_10(el, n):
    element = [el[n[0]-1], el[n[1]-1], el[n[2]-1], el[n[3]-1], el[n[4]-1],
               el[n[5]-1], el[n[6]-1], el[n[7]-1], el[n[8]-1], el[n[9]-1]]
    return element


#
#  辞書作成
neu2fistr_elDict = {
   #node数 要素名 node順      関数名
     3:["731", [1, 3, 2],    order_3],                      #三角形1次 (tri1)
     4:["341", [1, 3, 2, 4], order_4],                      #四面体1次 (tet1)
     6:["732", [1, 3, 2, 4, 6, 5], order_6],                #三角形2次 (tri2)
    10:["342", [1, 3, 2, 4, 8, 5, 6, 7, 10, 9], order_10]   #四面体2次 (tet2)
    }
fistr2neu_elDict = {
     3:["", [1, 3, 2],    order_3],                         #三角形1次 (tri1)
     4:["", [1, 3, 2, 4], order_4],                         #四面体1次 (tet1)
     6:["", [1, 3, 2, 4, 6, 5], order_6],                   #三角形2次 (tri2)
    10:["", [1, 3, 2, 4, 6, 7, 8, 5, 10, 9], order_10]      #四面体2次 (tet2)
    }
vol2fistr_elDict = neu2fistr_elDict
fistr2vol_eldict = fistr2neu_elDict
aba2neu_elDict = {
    3:["", [2, 3, 1], order_3],
    4:["", [1, 3, 2, 4], order_4],
   10:["", [1, 3, 2, 4, 6, 8, 5, 7, 10, 9], order_10]
   }

#
#  neu2fistr_el
#    neutral -> FrontISTR　変換
#    節点数から要素typeを判断し、node順を入れ替える
def neu2fistr_el(element):
    nNode = len(element)
    [name, n, func] = neu2fistr_elDict[nNode]
    newElement = func(element, n)
    return [name, newElement]

#
#  fistr2neu_el
#    FrontISTR -> neutral 変換
#    節点数から要素typeを判断し、node順を入れ替える
def fistr2neu_el(element):
    nNode = len(element)
    [name, n, func] = fistr2neu_elDict[nNode]
    newElement = func(element, n)
    return [name, newElement]

#
#  vol2fistr_el
#    vol -> FrontISTR 変換
#    節点数から要素typeを判断し、node順を入れ替える
def vol2fistr_el(element):
    return neu2fistr_el(element)

#
#  fistr2vol_el
#    FrontISTR -> vol 変換
#    節点数から要素typeを判断し、node順を入れ替える
def fistr2vol_el(element):
    return fistr2neu_el(element)

#  aba2neu_el
#    abaqus -> neutral 変換
#    節点数から要素typeを判断し、node順を入れ替える
def aba2neu_el(element):
    nNode = len(element)
    [name, n, func] = aba2neu_elDict[nNode]
    newElement = func(element, n)
    return [name, newElement]

#  aba2vol_el
#    abaqus -> vol 変換
def aba2vol_el(element):
    return aba2neu_el(element)

#  ----------------- FrontISTR専用 ----------------------------

faceElmDict = {
    "tet":[[1, [1,2,3]],
           [2, [1,2,4]],
           [3, [2,3,4]],
           [4, [3,1,4]]],
    "hex":[[1, [1,2,3,4]],
           [2, [5,6,7,8]],
           [3, [1,2,6,5]],
           [4, [2,3,7,6]],
           [5, [3,4,8,7]],
           [6, [4,1,5,8]]],
    "prs":[[1, [1,2,3]],
           [2, [4,5,6]],
           [3, [1,2,5,4]],
           [4, [2,3,6,5]],
           [5, [3,1,4,6]]]
    }

#
#  fistr_elFace
#    fistr要素のfaceNoを取得
#    （表面NodeNoから要素の面Noを取得）
def fistr_elFace(suNodes, elNodes):
    nEl = len(elNodes)
    if nEl == 4 or nEl == 10:
        #四面体
        suNodes = suNodes[:3]
        elNodes = elNodes[:4]
        suNodes.sort()
        faceNos = faceElmDict["tet"]
        for face in faceNos:
            faceNodes = order_3(elNodes, face[1])
            faceNodes.sort()
            if suNodes == faceNodes:
                faceNo = face[0]
                break
    elif nEl == 8 or nEl == 20:
        #六面体
        suNodes = suNodes[:4]
        elNodes = elNodes[:8]
        suNodes.sort()
        faceNos = faceElmDict("hex")
        for face in faceNoes:
            faceNodes = order_4(elNodes, face[1])
            faceNodes.sort()
            if suNodes == faceNodes:
                faceNo = face[0]
                break
    return faceNo


#---------------- neutral形式専用 -------------------------

#
#  createNeutralMeshLines
#    nodeList, surfaceList, elementListから、neutral形式のmeshLinesを作成
def createNeutralMeshLines(nodeList, surfaceList, elementList):

    def createLines(dataList):
        dataLines = []
        nMax = 0
        for data in dataList:
            if len(data) > 0:
                items = data[0]
                words = []
                for item in items:
                    words.append(str(item))
                line = " ".join(words) + ls
                dataLines += [line]
                nMax += 1
        lines = [str(nMax) + ls] + dataLines
        return lines

    #node行を作成
    lines = createLines(nodeList)
    #element行を作成
    lines += createLines(elementList)
    #surface行を作成
    lines += createLines(surfaceList)
    return lines    

#
#  getNeutralMeshConts
#    neutral形式のlinesから、nodeList, elementList, surfaceListを取得
def getNeutralMeshConts(lines):
    #各linesを取得
    i = 0
    nNode = int(lines[i])
    i += 1
    nodeLines = lines[i:i+nNode]
    i += nNode
    nElm = int(lines[i])
    i += 1
    elLines = lines[i:i+nElm]
    i += nElm
    nSur = int(lines[i])
    i += 1
    surLines = lines[i:i+nSur]

    #nodeList作成 nodeList[i] -> [[x,y,z], [eli, elj...], [sui, suj...]]
    #                             座標      要素No         surfaceNo
    #  要素No:    surfaceが属する要素の面番号を取得するため
    #  surfaceNo: 表面の節点を取得する為
    nodeList = [ [] for i1 in range(len(nodeLines) + 1) ]
    for i in range(len(nodeLines)):
        line = nodeLines[i]
        words = line.split()
        loc = [float(words[0]), float(words[1]), float(words[2])]
        nodeList[i+1] = [loc, [], []]
    #elementList作成 elementList -> [[grp, node1, node2,...]]
    elementList = [ [] for i1 in range(len(elLines) + 1) ]
    for i in range(len(elLines)):
        line = elLines[i]
        words = line.split()
        grp = words[0]
        nEl = len(words[1:])
        element = [[]] * (nEl+1)        #要素1ヶ分のlistを作成
        element[0] = int(grp)           #grpNo
        n = 1
        for word in words[1:]:
            nodeNo = int(word)
            element[n] = nodeNo
            nodeList[nodeNo][1] += [i+1]     #nodeListに要素Noを保存
            n += 1
        elementList[i+1] = [element]
    #surfaceList作成  surfaceList -> [[grp, node1, node2,...], [eli, elj...]]
    #                                   表面要素内容             要素No
    #  要素No: surfaceが属する面番号を取得するため
    surfaceList = [ [] for i1 in range(len(surLines) + 1) ]
    for i in range(len(surLines)):
        line = surLines[i]
        words = line.split()
        grp = words[0]
        nP = len(words[1:])
        surface = [[]] * (nP+1)
        surface[0] = int(grp)
        n = 1
        for word in words[1:]:
            nodeNo = int(word)
            surface[n] = nodeNo
            nodeList[nodeNo][2] += [i+1]   #nodeListにsurfaceNoを保存
            n += 1
        surfaceList[i+1] = [surface, []]
    #surfaceListに要素Noを追加
    for i in range(len(surfaceList)):
        surface = surfaceList[i]
        if len(surface) > 0:
            #nodeNoを取得
            nodes = surface[0][1:]
            if len(nodes) == 3 or len(nodes) == 6:
                #三角形
                nodes = nodes[:3]
            else:
                #四角形
                nodes = nodes[:4]
            #nodeから属する要素Noを取得
            elms = []
            for node in nodes:
                elms += nodeList[node][1]
            #同じ要素がnode数分ある要素を取得
            elms.sort()
            before = elms[0]
            n = 1
            for elm in elms[1:]:
                if elm == before:
                    n += 1
                else:
                    if n >= len(nodes):
                        break
                    n = 1
                    before = elm
            if n >= len(nodes):
                surfaceList[i][1] += [before]
            """
            #elms.sort()
            elms = deleteSameData(elms)
            #要素Noを取得
            for elNo in elms:
                elNodes = elementList[elNo][0][1:]
                #elNodes内にsurfaceのnode全てが含まれるelNoを抽出
                flag = 0
                for node in nodes:
                    if not (node in elNodes):
                        flag = 1
                        break
                if flag == 0:
                    #surfaceのnode全てがelm内に含まれる場合
                    surfaceList[i][1] += [elNo]
            """
    return [nodeList, elementList, surfaceList]

#--------------- stl 専用 -------------------------------------------

#
#  getStlNodeSurfaceEdgeList
#    stl三角形からnodeList、surfaceList、edgeListを作成する
#    入力:三角形座標のlist（[tri0],[tri1],[tri2], ...]
#                           tri = [loc0,loc1,loc2]
#                           loc = [x,y,z]
#    戻り:[nodeList, surfaceList, edgeList]
#         点：nodeList = [ [[x,y,z],[sur0,sur1...]], ...]
#         面：surfaceList = [ [[n0,n1,n2]], ...]
#         線：edgeList = [ [[n0, n1],[sur1,sur2...]], ...]
def getStlNodeSurfaceEdgeList(stlLocs):

    #  setNodeNos
    #    node座標を1列で取得
    def setNodeNos(stlLocs):
        locNos = [ [] for i in range(len(stlLocs) * 3) ]
        n = 0
        for locs in stlLocs:
            for loc in locs:
                locNos[n] = [loc]
                n += 1
        return locNos

    #  getMinMaxLoc
    #    minMaxの座標を取得
    def getMinMaxLoc(locNos):
        loc = locNos[0][0]
        minx = loc[0]; maxx = loc[0]
        miny = loc[1]; maxy = loc[1]
        minz = loc[2]; maxz = loc[2]
        for loc in locNos:
            x = loc[0][0]; y = loc[0][1]; z = loc[0][2]
            if minx > x:
                minx = x
            if miny > y:
                miny = y
            if minz > z:
                minz = z
            if maxx < x:
                maxx = x
            if maxy < y:
                maxy = y
            if maxz < z:
                maxz = z
        return [[minx, miny, minz], [maxx, maxy, maxz]]

    #  getLocAddress
    #    座標のaddressを取得
    def getLocAddress(loc, minMaxLoc):
        [x,y,z] = loc
        [[minx, miny, minz], [maxx, maxy, maxz]] = minMaxLoc
        ax = int((x - minx) / (maxx - minx) * 100.0 + 0.5)
        ay = int((y - miny) / (maxy - miny) * 100.0 + 0.5)
        az = int((z - minz) / (maxz - minz) * 100.0 + 0.5)
        return [ax, ay, az]    

    #  getNodeAddress
    #    node座標のaddressを作成
    def getNodeAddress(locNos, minMaxLoc):
        #3次元配列を作成
        nodeAdd = nodeAdd = [[[ [] for i3 in range(101)] for i2 in range(101)] for i1 in range(101)]
        #座標をaddressにセット
        [[minx, mixy, mixz], [maxx, maxy, maxz]] = minMaxLoc
        for i in range(len(locNos)):
            loc = locNos[i][0]
            [ax, ay, az] = getLocAddress(loc, minMaxLoc)
            nodeAdd[ax][ay][az] += [i]
        return nodeAdd

    #  getNewNodeNo
    #    nodeNoを取得
    def getNewNodeNo(locNos, locNo, nearLocs, tol, lastNodeNo):
        loc = locNos[locNo][0]
        #既に設定されているか
        if len(locNos[locNo]) > 1:
            #nodeNoを読み取り戻る
            nodeNo = locNos[locNo][1]
            return [nodeNo, locNos, lastNodeNo]
        #同じlocationのnodeNoを取得（近隣のnodeNoから）
        toll = geo.mag(loc) * tol + tol     #toleranceを設定
        sameLocNos = []
        for nearNo in nearLocs:
            nearLoc = locNos[nearNo][0]
            l = geo.length(loc, nearLoc)
            if l < toll:
                #同じlocationのnodeを保存
                sameLocNos.append(nearNo)
        #同じlocationでnodeNoが設定されていれば、読み取り戻る
        if len(locNos[sameLocNos[0]]) > 1:
            nodeNo = locNos[sameLocNos[0]][1]
            return [nodeNo, locNos, lastNodeNo]
        #新規にnodeNoを設定する
        lastNodeNo += 1
        nodeNo = lastNodeNo
        #同じlocation内のnodeにnodeNoを設定
        for sameLocNo in sameLocNos:
            if len(locNos[sameLocNo]) > 1:
                locNos[sameLocNo][1] = nodeNo
            else:
                locNos[sameLocNo] += [nodeNo]
        return [-1, locNos, lastNodeNo]

    #  getEdgeNo
    #    edgeAddressからedgeNoを取得
    def getEdgeNo(edge, edgeList, edgeAdd):
        if edgeAdd[edge[0]][edge[1]] != 0:
            edgeNo = edgeAdd[edge[0]][edge[1]]
        elif edgeAdd[edge[1]][edge[0]] != 0:
            edgeNo = edgeAdd[edge[1]][edge[0]]
        else:
            edgeNo = len(edgeList)
            edgeList.append([edge, []])
            edgeAdd[edge[0]][edge[1]] = edgeNo
        return [edgeNo, edgeList, edgeAdd]

    #
    #toleranceを設定（有効数字6桁）
    tol = 1.0e-6

    #nodeList, surfaceListを作成
    #  minMaxLocを取得
    locNos = setNodeNos(stlLocs)
    minMaxLoc = getMinMaxLoc(locNos)
    #  nodeAddを作成
    nodeAdd = getNodeAddress(locNos, minMaxLoc)
    #  nodeNoを取得
    nodeList = [[]]
    surfaceList = [ [] for i in range(len(stlLocs)+1)]
    lastNodeNo = 0
    n = 0
    for i in range(len(stlLocs)):
        surNo = i + 1
        locs = stlLocs[i]
        surNodeNos = []
        for loc in locs:
            [ax, ay, az] = getLocAddress(loc, minMaxLoc)
            nearLocs = nodeAdd[ax][ay][az]
            #nodeNoを取得、locNos[i]の最後にnodeNoを書き込む、lastNodeを更新
            #新規にnodeNoを追加した場合、nodeNoは「-1」が戻る
            [nodeNo, locNos, lastNodeNo] = getNewNodeNo(locNos, n, nearLocs, tol, lastNodeNo)
            if nodeNo < 0:
                #新規にnodeNoが設定された場合
                nodeNo = lastNodeNo
                nodeList.append([loc, []])
            nodeList[nodeNo][1] += [surNo]
            surNodeNos.append(nodeNo)
            n += 1
        surfaceList[surNo] = [surNodeNos]
    #メモリクリア
    locNos = []
    nodeAdd = []

    #edgeListを作成
    #  edgeのaddress用配列を作成
    edgeAdd = [[0 for i1 in range(len(nodeList)+1)] for i2 in range(len(nodeList)+1)]
    #  最初のedgeを設定
    edge0 = [surfaceList[1][0][0], surfaceList[1][0][1]]
    edgeAdd[edge0[0]][edge0[1]] = 1
    edgeList = [[], [edge0, []]]
    #  2番目以降のedgeListを作成
    for i in range(1, len(surfaceList)):
        surface = surfaceList[i]
        surNo = i
        edge = [surface[0][0], surface[0][1]]
        [edgeNo, edgeList, edgeAdd] = getEdgeNo(edge, edgeList, edgeAdd)
        edgeList[edgeNo][1] += [surNo]
        edge = [surface[0][1], surface[0][2]]
        [edgeNo, edgeList, edgeAdd] = getEdgeNo(edge, edgeList, edgeAdd)
        edgeList[edgeNo][1] += [surNo]
        edge = [surface[0][2], surface[0][0]]
        [edgeNo, edgeList, edgeAdd] = getEdgeNo(edge, edgeList, edgeAdd)
        edgeList[edgeNo][1] += [surNo]
    return [nodeList, surfaceList, edgeList]        
        
        
