#!/usr/bin/env python
# -*- coding: utf-8 -*-    
#
#   mergeMesh.py
#
#       分割されたneutral形式のpartMeshメッシュをmergeする
#
#   18/02/22    新規作成
#

import sys
import glob
from multiprocessing import Process, Queue
import convertMesh as cm
import getGroupsFromVolStl as getGrp

ls = "\n"


#
#  getProcsHeaderConts
#    procsの内容をheader毎に取得
def getProcsHeaderConts(lines):
    headerConts = []
    flag = 0
    headerCont = ["", []]
    for line in lines:
        if line[:len("# *")] == "# *":
            if headerCont[0] != "":
                headerConts.append(headerCont)
            headerCont = ["", []]
            header = line[2:]
            headerCont[0] = header
            flag = 1
        else:
            if flag == 1:
                headerCont[1].append(line[2:])
    if headerCont[0] != "":
        headerConts.append(headerCont)
    return headerConts

#
#  getGrpsProcsData
#    各groupとprocDataを取得
def getGrpsProcsData(headerConts, elementList, surfaceList):
    #elementGroupを取得
    grps = []
    for headerCont in headerConts:
        header = headerCont[0]
        words = header.split()
        if words[0] == "*elementGroup":
            grpNo = int(words[1])
            grpName = words[2]
            grps.append([grpNo, grpName])
    elmGrps = []
    for grp in grps:
        grpNo = grp[0]
        grpName = grp[1]
        elms = [grpName, []]
        i = 0
        for elmData in elementList:
            if len(elmData) > 0:
                if elmData[0][0] == grpNo:
                    elms[1].append(i)
            i += 1
        elmGrps.append(elms)
    #surfaceGroupを取得
    grps = []
    for headerCont in headerConts:
        header = headerCont[0]
        words = header.split()
        if words[0] == "*surfaceGroup":
            grpNo = int(words[1])
            grpName = words[2]
            grps.append([grpNo, grpName])
    surGrps = []
    for grp in grps:
        grpNo = grp[0]
        grpName = grp[1]
        surs = [grpName, []]
        data = []
        for i in range(len(surfaceList)):
            if len(surfaceList[i]) > 0:
                if grpNo == surfaceList[i][0][0]:
                    surs[1].append(i)
        surGrps.append(surs)
    #nodeGroupを取得
    nodeGrps = []
    for headerCont in headerConts:
        header = headerCont[0]
        words = header.split()
        if words[0] == "*nodeGroup":
            grpName = words[1]
            grp = [grpName, []]
            grpLines = headerCont[1]
            for line in grpLines:
                words = line.split()
                for val in words:
                    grp[1] += [int(val)]
            nodeGrps.append(grp)
    #procDataを取得
    procData = []
    for headerCont in headerConts:
        header = headerCont[0]
        words = header.split()
        if words[0][:len("*toProc_")] == "*toProc_":
            procName = words[0]
            proc = [procName, []]
            pData = headerCont[1]
            for line in pData:
                words = line.split()
                for val in words:
                    ww = val.split(":")
                    proc[1] += [[int(ww[0]), int(ww[1])]]
            procData.append(proc)
    #dataCombine
    grps = [nodeGrps, elmGrps, surGrps]
    return [grps, procData]

#
#  getInternalSurfaces
#    internalSurfaceNoを取得
def getInternalSurfaces(headerConts):
    intSurNo = []
    for headerCont in headerConts:
        header = headerCont[0]
        words = header.split()
        if words[0] == "*internalSurfaceNo":
            pData = headerCont[1]
            for line in pData:
                words = line.split()
                for word in words:
                    surNo = int(word)
                    intSurNo.append(surNo)
    return intSurNo

#
#  readProcMeshFile
#    procのメッシュfileを読み込む
def readProcMeshFile(meshFile):
    procNo = int(meshFile.split(".")[-2])
    lines = getGrp.readFileLines(meshFile)
    [nodeList, elementList, surfaceList] = cm.getNeutralMeshConts(lines)
    p = len(nodeList) + len(elementList) + len(surfaceList)
    headerConts = getProcsHeaderConts(lines[p:])
    [neshGrps, procs] = getGrpsProcsData(headerConts, elementList, surfaceList)
    internalSurs = getInternalSurfaces(headerConts)
    nesList = [nodeList, elementList, surfaceList]
    meshConts = [procNo, [nesList, neshGrps, procs]]
    return [meshConts, internalSurs]

#
#  remakeElmSurNoInNodeList
#    nList内のelmNo、surNoを修正
def remakeElmSurNoInNodeList(lastElmNo, lastSurNo, nList):
    for i in range(1, len(nList)):
        #elmNoを修正
        elms = nList[i][1]
        for j in range(len(elms)):
            elmNo = elms[j]
            nList[i][1][j] += lastElmNo
        #surNoを修正
        for j in range(len(nList[i][2])):
            nList[i][2][j] += lastSurNo
    return nList

#
#  addNewNodeNo
#    nodeを追加してnewNodeNoを設定
def addNewNodeNo(nodeList, nList):
    firstNo = len(nodeList)
    for i in range(1, len(nList)):
        newNo = firstNo + i - 1
        nList[i] += [newNo]
    return nList

#
#  setNewNodeNo
#    nodeListをmergeする
def setNewNodeNo(nodeList, nList, myProcs, myProcNo, addProcNo):
    #共有させるnodeを取得
    mergeNodes = []
    for procCont in myProcs:
        procNo = int(procCont[0].split("_")[-1])
        if procNo == addProcNo:
            mergeNodes = procCont[1]
            break
    #mergeする側のnodeNoを修正
    #  共有nodeNoを書き込み
    for pairNodes in mergeNodes:
        myNode = pairNodes[0]
        addNode = pairNodes[1]
        nList[addNode] += [myNode]
    #  新nodeNoを書き込み、newList作成
    newList = []
    maxNodeNo = len(nodeList)
    n = maxNodeNo
    for i in range(len(nList)):
        if len(nList[i]) > 0:
            if len(nList[i]) == 1:
                nList[i] += [n]
                newList.append(nList[i])
                n += 1
    return [nList, newList]

#
#  remakeElementNodeNo
#    elementのnodeNoを修正
def remakeElementNodeNo(lastNo, eList):
    for i in range(len(eList)):
        if len(eList[i]) > 0:
            conts = eList[i][0]
            for j in range(1, len(conts)):
                nodeNo = conts[j]
                eList[i][0][j] += lastNo
    return eList

#
#  remakeElementNoInSurfaceList
#    sList内の要素Noを修正
def remakeElementNoInSurfaceList(lastElmNo, sList):
    for i in range(len(sList)):
        if len(sList[i]) > 0:
            for j in range(len(sList[i][1])):
                sList[i][1][j] += lastElmNo
    return sList

#
#  remakeNodeGrpsNodeNo
#    nodeGrpsのNodeNoを修正
def remakeNodeGrpsNodeNo(nList, nGrps):
    for i in range(len(nGrps)):
        grpName = nGrps[i][0]
        grpData = nGrps[i][1]
        for j in range(len(grpData)):
            nodeNo = grpData[j]
            newNo = nList[nodeNo][-1]
            nGrps[i][1][j] = newNo
    return nGrps

#
#  remakeElmGrpsElmNo
#    eGrpsの要素Noを修正
def remakeElmGrpsElmNo(lastNo, eGrps):
    for i in range(len(eGrps)):
        name = eGrps[i][0]
        eData = eGrps[i][1]
        for j in range(len(eData)):
            eGrps[i][1][j] += lastNo
    return eGrps

#
#  addGrps
#    同じgrpがあればそれをを加える
def addGrps(nodeGrps, nGrps):

    def addGrp(grp, nodeGrps):
        flag = 0
        for i in range(len(nodeGrps)):
            if grp[0] == nodeGrps[i][0]:
                nodeGrps[i][1] += grp[1]
                flag = 1
        if flag == 0:
            nodeGrps += [grp]
        return nodeGrps

    for nGrp in nGrps:
        nodeGrps = addGrp(nGrp, nodeGrps)
    return nodeGrps

#
#  addMeshConts
#    meshを追加する
def addMeshConts(meshConts, pMeshConts):
    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    [addProcNo, [pNesList, pNesGrps, procs]] = pMeshConts
    [nList, eList, sList] = pNesList
    [nGrps, eGrps, sGrps] = pNesGrps
    #各listの最終Noを取得（この値分シフトさせる）
    lastNodeNo = len(nodeList) - 1
    lastElmNo = len(elementList) - 1
    lastSurNo = len(surfaceList) - 1
    #nList内のelementNo、surfaceNoを修正
    nList = remakeElmSurNoInNodeList(lastElmNo, lastSurNo, nList)
    #eList, sListのnodeNoを修正
    eList = remakeElementNodeNo(lastNodeNo, eList)
    sList = remakeElementNodeNo(lastNodeNo, sList)
    #sList内のelementNoを修正
    sList = remakeElementNoInSurfaceList(lastElmNo, sList)
    #eGrps, sGrpsのnodeNoを修正
    eGrps = remakeElmGrpsElmNo(lastElmNo, eGrps)
    sGrps = remakeElmGrpsElmNo(lastSurNo, sGrps)
    nGrps = remakeElmGrpsElmNo(lastNodeNo, nGrps)
    #addMesh
    nodeList += nList[1:]
    elementList += eList[1:]
    surfaceList += sList[1:]
    elementGrps = addGrps(elementGrps, eGrps)
    surfaceGrps = addGrps(surfaceGrps, sGrps)
    nodeGrps = addGrps(nodeGrps, nGrps)
    #meshをcombine
    nesList = [nodeList, elementList, surfaceList]
    nesGrps = [nodeGrps, elementGrps, surfaceGrps]
    meshConts = [myProcNo, [nesList, nesGrps, myProcs]]
    return meshConts

#
#  createFistrMesh
#    FrontISTR用のメッシュを作成
def createFistrMesh(nesList, nesGrps):
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    lines = []
    #node行を作成
    lines += ["!NODE" + ls]
    p = 1
    for i in range(p, len(nodeList)):
        loc = nodeList[i][0]
        words = [str(i), str(loc[0]), str(loc[1]), str(loc[2])]
        line = ", ".join(words) + ls
        lines += [line]
    #element行を作成
    for grp in elementGrps:
        name = grp[0]
        elData = grp[1]
        elNo = elData[0]
        elNodes = elementList[elNo][0][1:]
        [elType, nodes] = cm.neu2fistr_el(elNodes)
        header = "!ELEMENT, TYPE=" + elType + ", EGRP=" + name + ls
        lines += [header]
        for elNo in elData:
            element = elementList[elNo][0][1:]
            [dummy, elNodes] = cm.neu2fistr_el(element)
            words = [str(elNo)]
            for el in elNodes:
                words.append(str(el))
            line = ", ".join(words) + ls
            lines += [line]
        lines += ["!SECTION, TYPE=SOLID, EGRP=" + name + ls]
    #surfaceGroupを作成
    for surGroup in surfaceGrps:
        surName = surGroup[0]
        surData = surGroup[1]
        header = "!SGROUP, SGRP=" + surName + ls
        lines += [header]
        newSurData = []
        for surNo in surData:
            surNodes = surfaceList[surNo][0][1:]
            surElm = surfaceList[surNo][1][0]
            elNodes = elementList[surElm][0][1:]
            [name, surNodes] = cm.vol2fistr_el(surNodes)
            [name, elNodes] = cm.vol2fistr_el(elNodes)
            faceNo = cm.fistr_elFace(surNodes, elNodes)
            newSurData.append([surElm, faceNo])
        newSurData.sort()
        for data in newSurData:
            line = str(data[0]) + ", " + str(data[1]) + ls
            lines += [line]
    #nodeGroupを作成
    for grp in nodeGrps:
        name = grp[0]
        data = grp[1]
        header = "!NGROUP, NGRP=" + name + ls
        lines += [header]
        n = 10
        words = []
        for node in data:
            words.append(str(node))
            n -= 1
            if n <= 0:
                line = ", ".join(words) + ", " + ls
                lines += [line]
                n = 10
                words = []
        if len(words) != 0:
            line = ", ".join(words) + ls
            lines += [line]
    lines += ["!END" + ls]
    return lines

#
#  fuseSeparateSurface
#    分割面を結合する
def fuseSeparateSurface(procSepNodes, meshConts, nProcNodes):

    def getLastNodeNo(procNo, nProcNodes):
        lastNo = 0
        for i in range(procNo):
            if i < len(nProcNodes):
                lastNo += nProcNodes[i]
            else:
                break
        return lastNo

    #変数を展開
    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    #pairNodeを取得（nodeを修正）
    pairNodes = []
    for proc0 in range(len(procSepNodes)):
        for j in range(len(procSepNodes[proc0])):
            name = procSepNodes[proc0][j][0]
            proc1 = int(name.split("_")[-1])
            #proc0-proc1間をチェック
            lastNo0 = getLastNodeNo(proc0, nProcNodes)
            lastNo1 = getLastNodeNo(proc1, nProcNodes)
            if lastNo1 < len(nodeList)-1:
                #pairを取得
                comNodes = procSepNodes[proc0][j][1]
                for comNode in comNodes:
                    node0 = comNode[0] + lastNo0
                    node1 = comNode[1] + lastNo1
                    pairNodes.append([node0, node1])
    #nodeGrpListを作成
    nodeGrpList = [ [] for i in range(len(nodeList)+1) ]
    for i in range(len(nodeGrps)):
        nodeData = nodeGrps[i][1]
        for node in nodeData:
            nodeGrpList[node] += [i]
    #結合開始
    for pair in pairNodes:
        node0 = pair[0]
        node1 = pair[1]
        #node1をnode0に書き換える
        #  elementListを修正
        elements = nodeList[node1][1]
        for elmNo in elements:
            elCont = elementList[elmNo][0]
            for i in range(1, len(elCont)):
                if elCont[i] == node1:
                    #nodeNoを書き換え
                    elementList[elmNo][0][i] = node0
                    break
        #  surfaceListを修正
        surfaces = nodeList[node1][2]
        for surNo in surfaces:
            surCont = surfaceList[surNo][0]
            for i in range(1, len(surCont)):
                if surCont[i] == node1:
                    #nodeNoを書き換え
                    surfaceList[surNo][0][i] = node0
                    break
        #  削除するnode(node1)にマーク付け
        nodeList[node1].append(-1)
        #  nodeGrpを確認し、残すnodeと削除するnodeのnodeGroupを合わせる
        nGrp0 = nodeGrpList[node0]
        nGrp1 = nodeGrpList[node1]
        nGrp0.sort()
        nGrp1.sort()
        if nGrp0 != nGrp1:
            #nodeGroupが異なっているので、合わせる
            nGrp = nGrp0 + nGrp1
            nGrp = getGrp.deleteSameData(nGrp)
            #追加するnodeGrpNoを取得
            addGrp = []
            for grp in nGrp:
                flag = 0
                for g in nGrp0:
                    if grp == g:
                        flag = 1
                        break
                if flag == 0:
                    addGrp.append(grp)
            #足りないnodeGrpを追加
            for grp in addGrp:
                nodeGrps[grp][1] += [node0]
    #meshをcombine
    nesList = [nodeList, elementList, surfaceList]
    nesGrps = [nodeGrps, elementGrps, surfaceGrps]
    meshConts = [myProcNo, [nesList, nesGrps, myProcs]]
    return meshConts

#
#  renumberNodes(meshConts)
#    余分なnodeを削除し、renumberingする
def renumberNodes(meshConts):
    #変数を展開
    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    #newNoを設定
    newNo = 1
    for i in range(1, len(nodeList)):
        if len(nodeList[i]) < 4:
            nodeList[i] += [newNo]
            newNo += 1
    #newNodeListを作成
    newNodeList = []
    for cont in nodeList:
        if len(cont) == 0:
            newNodeList.append(cont)
        else:
            if cont[-1] > 0:
                newNodeList.append(cont[:-1])
    #elementListのnodeNoを書き換え
    for i in range(1, len(elementList)):
        nodeCont = elementList[i][0]
        for j in range(1, len(nodeCont)):
            nodeNo = elementList[i][0][j]
            newNo = nodeList[nodeNo][-1]
            if newNo > 0:
                elementList[i][0][j] = newNo
    #surfaceListのnodeNoを書き換え
    for i in range(1, len(surfaceList)):
        nodeCont = surfaceList[i][0]
        for j in range(1, len(nodeCont)):
            nodeNo = surfaceList[i][0][j]
            newNo = nodeList[nodeNo][-1]
            if newNo > 0:
                surfaceList[i][0][j] = newNo
    #nodeGrpのnodeNoを書き換え
    for i in range(len(nodeGrps)):
        name = nodeGrps[i][0]
        nodeData = nodeGrps[i][1]
        newData = []
        for j in range(len(nodeData)):
            nodeNo = nodeGrps[i][1][j]
            newNo = nodeList[nodeNo][-1]
            if newNo > 0:
                newData.append(newNo)
        nodeGrps[i][1] = newData
    #meshをcombine
    nesList = [newNodeList, elementList, surfaceList]
    nesGrps = [nodeGrps, elementGrps, surfaceGrps]
    meshConts = [myProcNo, [nesList, nesGrps, myProcs]]
    return meshConts

#
#  sortGroupName(meshConts)
#    各groupを名前順にsort
def sortGroupName(meshConts):

    #Grpsをsortする
    def sortGroup_sub(grps, grpName):
        for i in range(len(grps)):
            name = grps[i][0]
            if name[:len(grpName)] == grpName:
                n = int(name.split("_")[-1])
                grps[i][0] = n
        grps.sort()
        for i in range(len(grps)):
            name = grps[i][0]
            if type(name) == int:
                name = grpName + str(name)
                grps[i][0] = name
        return grps

    #変数を展開
    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    #nodeGrpsのsort
    nodeGrps = sortGroup_sub(nodeGrps, "node_")
    #surfaceGrpsのsort
    surfaceGrps = sortGroup_sub(surfaceGrps, "surface_")
    #elementGrpsのsort
    elementGrps = sortGroup_sub(elementGrps, "elm_")
    #meshをcombine
    nesList = [nodeList, elementList, surfaceList]
    nesGrps = [nodeGrps, elementGrps, surfaceGrps]
    meshConts = [myProcNo, [nesList, nesGrps, myProcs]]
    return meshConts    

#
#  printMeshConts
#    mesh内容を出力
def printMeshConts(meshConts):
    #変数を展開
    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    print "  nodes: " + str(len(nodeList)-1) + "  elements: " + str(len(elementList)-1) + "  surfaces: " + str(len(surfaceList)-1)
    print "  node groups"
    for nodeGrp in nodeGrps:
        name = nodeGrp[0]
        data = nodeGrp[1]
        print "    " + name + ": " + str(len(data))
    print "  element groups"
    for elmGrp in elementGrps:
        name = elmGrp[0]
        data = elmGrp[1]
        print "    " + name + ": " + str(len(data))
    print "  surface groups"
    for grp in surfaceGrps:
        name = grp[0]
        data = grp[1]
        print "    " + name + ": " + str(len(data))

#
#  processReadProcMesh
#    procMesh読み込み用のprocess
def processReadProcMesh(procFile, q):
    procNo = int(procFile.split(".")[-2])
    [meshConts, dummy] = readProcMeshFile(procFile)
    q.put(meshConts)

#
#  mergeMesh
#  ---------
def mergeMesh(fileHeader):
    print ls + "Merge Mesh"
    #procFilesを取得
    procFiles = glob.glob(fileHeader + ".*.mesh")
    maxProcNo = len(procFiles) - 1

    #各processを起動し、各procMeshを読み込む
    #  processを作成（proc1以降のprocessを実行）
    processes = []
    queues = []
    for i in range(1, len(procFiles)):
        procFile = fileHeader + "." + str(i) + ".mesh"
        q = Queue()
        p = Process(target=processReadProcMesh, args=(procFile, q,))
        processes.append(p)
        queues.append(q)
        p.start()

    #最初のprocを読み込む
    volFile = fileHeader + ".0.mesh"
    print "reading " + volFile + "..."
    [meshConts, intSurs] = readProcMeshFile(volFile)
    sepNodes = [meshConts[1][2]]
    nProcNodes = [len(meshConts[1][0][0]) - 1]

    #proc毎にmergeする
    for i in range(1, len(procFiles)):
        print "adding " + fileHeader + "." + str(i) + ".mesh..."
        #各processの結果を取得
        pNo = i - 1
        procMeshConts = queues[pNo].get()
        #各meshを加える
        sepNodes += [procMeshConts[1][2]]
        nProcNodes += [len(procMeshConts[1][0][0]) - 1]
        meshConts = addMeshConts(meshConts, procMeshConts)

    print "fusing separated surfaces of each proccessor..."
    meshConts = fuseSeparateSurface(sepNodes, meshConts, nProcNodes)
    print "renumbering node number..."
    meshConts = renumberNodes(meshConts)
    meshConts = sortGroupName(meshConts)
    printMeshConts(meshConts)

    [myProcNo, [nesList, nesGrps, myProcs]] = meshConts
    [nodeList, elementList, surfaceList] = nesList
    [nodeGrps, elementGrps, surfaceGrps] = nesGrps
    print "writing mesh for FrontISTR format..."
    lines = createFistrMesh(nesList, nesGrps)
    getGrp.writeFileLines("FistrModel.msh", lines)
    print "  ... done"

if __name__ == "__main__":
    import gettext
    gettext.install("app")
    
    fileHeader = sys.argv[1]
    mergeMesh(fileHeader)

