#!/usr/bin/env python  
# -*- coding: utf-8 -*-    
#
#   setGroupsFromPreMesh.py
#
#       preMeshからGroupを取得し、各々のvol形式メッシュに設定する。
#
#   18/02/11    新規作成
#
#

import sys
import os
import geometryFunc as geo
import convertMesh as cm
import getGroupsFromVolStl as getvol

ls = "\n"


#
#  separateGroups
#    group別に分ける
def separateGroups(grpData):
    #grpNoを取得
    grpNos = []
    for data in grpData:
        if len(data) > 0:
            grp = data[0]
            grpNos.append(grp)
    grpNos = getvol.deleteSameData(grpNos)
    #索引テーブル作成
    maxNo = grpNos[-1]
    grpIdx = [[]] * (maxNo + 1)
    for i in range(len(grpNos)):
        grpNo = grpNos[i]
        grpIdx[grpNo] = i
    #groupデータを取得
    grps = []
    for i in range(len(grpNos)):
        grps.append([grpNos[i], []])
    for listData in grpData:
        if len(listData) > 0:
            grpNo = listData[0]
            data = listData[1]
            idx = grpIdx[grpNo]
            grps[idx][1] += [data]
    return grps

#
#  getTriTetFromPreMesh
#    preMesh内のtri、tetraを取得
def getTriTetFromPreMesh(nodeList, elementList, surfaceList):
    #triを取得（表面三角形）
    tri = [[]] * (len(surfaceList) + 1)
    maxGrp = -1
    for i in range(len(surfaceList)):
        if len(surfaceList[i]) > 0:
            grp = surfaceList[i][0][0]
            if maxGrp < grp:
                maxGrp = grp
            nodes = surfaceList[i][0][1:4]
            loc1 = nodeList[nodes[0]][0]
            loc2 = nodeList[nodes[1]][0]
            loc3 = nodeList[nodes[2]][0]
            tri[i] = [grp, [loc1, loc2, loc3]]
    tri = separateGroups(tri)
    #tetを取得（要素）
    tet = [[]] * (len(elementList) + 1)
    for i in range(len(elementList)):
        if len(elementList[i]) > 0:
            grp = elementList[i][0][0]
            nodes = elementList[i][0][1:5]
            loc1 = nodeList[nodes[0]][0]
            loc2 = nodeList[nodes[1]][0]
            loc3 = nodeList[nodes[2]][0]
            loc4 = nodeList[nodes[3]][0]
            tet[i] = [grp, [loc1, loc2, loc3, loc4]]
    tet = separateGroups(tet)
    return [tri, tet]

#
#  getSurfaceGrpNoOfElmGrps
#    elementGrp間境界面のsurfaceGrpNoを取得
def getSurfaceGrpNoOfElmGrps(surfaceList):
    elSep = []
    for surfaceCont in surfaceList:
        if len(surfaceCont) > 0:
            if len(surfaceCont[1]) > 1:
                surNo = surfaceCont[0][0]
                elSep.append(surNo)
    elSep = getvol.deleteSameData(elSep)
    return elSep

#
#  readPreMeshConts
#    preMeshの内容を取得
def readPreMeshConts(preMeshFile):
    #premeshの読み込み
    [nodeList, elementList, surfaceList, preNodeGrps] = getvol.readVolMesh(preMeshFile)
    #表面三角形と四面体の座標を取得
    [preTri, preTet] = getTriTetFromPreMesh(nodeList, elementList, surfaceList)
    #elementGrp間の内部surfaceNoを取得
    elSep = getSurfaceGrpNoOfElmGrps(surfaceList)
    preMeshList = [nodeList, elementList, surfaceList]
    return [preMeshList, preTri, preTet, elSep, preNodeGrps]

#
#  setSurfaceGroups
#    surfaceListにGroupNoをセット
def setSurfaceGroups(nodeList, surfaceList, preLocs, minMaxLoc):
    #表面要素のaddressを作成
    surAdd = setSurfaceAddress(nodeList, surfaceList, minMaxLoc)
    #全てのfaceにデフォルトのfaceNoを設定（デフォルトのfaceNoは、分割面になる）
    maxNo = preLocs[-1][0]
    for i in range(len(surfaceList)):
        if len(surfaceList[i]) > 0:
            surfaceList[i][0][0] = maxNo + 1    #faceNoを設定
    #設定or未設定用のflag
    elmsFlag = [0] * len(surfaceList)
    #preGroupのloop
    for preData in preLocs:
        grpNo = preData[0]
        preTris = preData[1]
        #preGroupのloop
        for preTri in preTris:
            #blockを取得
            xx = [preTri[0][0], preTri[1][0], preTri[2][0]]
            yy = [preTri[0][1], preTri[1][1], preTri[2][1]]
            zz = [preTri[0][2], preTri[1][2], preTri[2][2]]
            minx = min(xx); maxx = max(xx)
            miny = min(yy); maxy = max(yy)
            minz = min(zz); maxz = max(zz)
            [ax, ay, az] = getLocAddress([minx, miny, minz], minMaxLoc)
            [bx, by, bz] = getLocAddress([maxx, maxy, maxz], minMaxLoc)
            #ax, bxを調整する
            [ax, ay, az] = fitIndexA([ax, ay, az])
            [bx, by, bz] = fitIndexB([bx, by, bz])
            #[ax, ay, az] = fitIndexA([ax, ay, az])
            #[bx, by, bz] = fitIndexB([bx, by, bz])
            #block内addressを取得
            blockSur = getLocAddInBlock(surAdd, [ax, ay, az], [bx, by, bz])
            #block内のloop
            for elmNo in blockSur:
                #設定or未設定
                if elmsFlag[elmNo] == 0:
                    #faceの角度をチェック
                    nodes = surfaceList[elmNo][0][1:4]
                    loc0 = nodeList[nodes[0]][0]
                    loc1 = nodeList[nodes[1]][0]
                    loc2 = nodeList[nodes[2]][0]
                    triLoc = [loc0, loc1, loc2]
                    cosAng = geo.facesAngle(preTri, triLoc)
                    if cosAng > 0.9659:
                        #face重心との距離チェック
                        centerLoc = geo.triG(triLoc)        #face重心
                        l = geo.isPointInFace(preTri, centerLoc)
                        tol = geo.length(preTri[0], preTri[1]) * 0.5
                        if 0.0 <= l and l < tol:
                            #grpNoとflagを設定
                            surfaceList[elmNo][0][0] = grpNo
                            elmsFlag[elmNo] = 1
                        """
                        [flg, l] = geo.isPointInFaceLarge(preTri, centerLoc)
                        if flg >= 0:
                            tol = geo.length(preTri[0], preTri[1]) * 0.5
                            #tol = geo.length(loc0, loc1) / 2.0
                            if 0.0 <= l and l < tol:
                                #grpNoとflagを設定
                                surfaceList[elmNo][0][0] = grpNo
                                if flg == 0:
                                    elmsFlag[elmNo] = 1
                        """
    return surfaceList

#
#  getLocAddress
#    座標のaddressを取得
def getLocAddress(XYZ, minMaxXYZ):
    [x,y,z] = XYZ
    [[minx, miny, minz], [maxx, maxy, maxz]] = minMaxXYZ
    ax = int((x - minx) / (maxx - minx) * 100.0 + 0.5)
    ay = int((y - miny) / (maxy - miny) * 100.0 + 0.5)
    az = int((z - minz) / (maxz - minz) * 100.0 + 0.5)
    return [ax, ay, az]    

#
#  getMinMaxLocation
#    最大最小の座標を取得する
def getMinMaxLocation(nodeList):
    [x, y, z] = nodeList[1][0]
    minx = x; miny = y; minz = z
    maxx = x; maxy = y; maxz = z
    i = 1
    while i < len(nodeList):
        [x,y,z] = nodeList[i][0]
        if minx > x:
            minx = x
        if miny > y:
            miny = y
        if minz > z:
            minz = z
        if maxx < x:
            maxx = x
        if maxy < y:
            maxy = y
        if maxz < z:
            maxz = z
        i += 1
    #minMax値を1%上乗せ
    dx = (maxx - minx) / 100.0
    dy = (maxy - miny) / 100.0
    dz = (maxz - minz) / 100.0
    minx = minx - dx; maxx = maxx + dx
    miny = miny - dy; maxy = maxy + dy
    minz = minz - dz; maxz = maxz + dz
    minxyz = [minx, miny, minz]
    maxxyz = [maxx, maxy, maxz]
    return [minxyz, maxxyz]    

#
#  getLocAddInBlock
#    block内のlocAddを取得
def getLocAddInBlock(allAdds, axyz, bxyz):
    [ax, ay, az] = axyz
    [bx, by, bz] = bxyz
    adds = []
    for x in range(ax, bx+1):
        if 0 <= x and x <= 100:
            for y in range(ay, by+1):
                if 0 <= y and y <= 100:
                    for z in range(az, bz+1):
                        if 0 <= z and z <= 100:
                            adds += allAdds[x][y][z]
    return adds

#
#  setNodeAddress
#    nodeGrpの位置addressを設定
def setNodeAddress(nodeList, nodeGrp, minMaxLoc):
    [[minx, miny, minz], [maxx, maxy, maxz]] = minMaxLoc
    #3次元リスト作成
    nodeAdd = [[[ [] for i3 in range(101)] for i2 in range(101)] for i1 in range(101)]
    #addressを作成
    for nodeNo in nodeGrp:
        loc = nodeList[nodeNo][0]
        [ax, ay, az] = getLocAddress(loc, minMaxLoc)
        nodeAdd[ax][ay][az] += [nodeNo]
    return nodeAdd    

#
#  setSurfaceAddress
#    表面要素の位置addressを設定
def setSurfaceAddress(nodeList, surfaceList, minMaxLoc):
    [[minx, miny, minz], [maxx, maxy, maxz]] = minMaxLoc
    #3次元リスト作成
    surAdd = [[[[] for i3 in range(101)] for i2 in range(101)] for i1 in range(101)]
    #addressを作成
    for i in range(len(surfaceList)):
        if len(surfaceList[i]) > 0:
            #重心を取得
            nodes = surfaceList[i][0][1:4]
            loc0 = nodeList[nodes[0]][0]
            loc1 = nodeList[nodes[1]][0]
            loc2 = nodeList[nodes[2]][0]
            tri = [loc0, loc1, loc2]
            [x,y,z] = geo.triG(tri)
            #重心のaddressを取得
            [ax, ay, az] = getLocAddress([x,y,z], minMaxLoc)
            surAdd[ax][ay][az] += [i]
    return surAdd      

#
#  setElementAddress
#    要素の位置adressを設定
def setElementAddress(nodeList, elementList, minMaxLoc):
    [minxyz, maxxyz] = minMaxLoc
    [minx, miny, minz] = minxyz
    [maxx, maxy, maxz] = maxxyz
    #3次元リストを作成
    elmAdd = [[[[] for i3 in range(101)] for i2 in range(101)] for i1 in range(101)]
    #addressを作成
    for i in range(len(elementList)):
        if len(elementList[i]) > 0:
            #重心を取得
            nodes = elementList[i][0][1:5]
            loc0 = nodeList[nodes[0]][0]
            loc1 = nodeList[nodes[1]][0]
            loc2 = nodeList[nodes[2]][0]
            loc3 = nodeList[nodes[3]][0]
            tetra = [loc0, loc1, loc2, loc3]
            [x,y,z] = geo.tetG(tetra)           #重心
            #重心のaddressを計算
            [ax, ay, az] = getLocAddress([x,y,z], minMaxLoc)
            #要素Noをaddressに追加して保存
            elmAdd[ax][ay][az] += [i]
    return elmAdd

#  fitIndexA
#    index範囲を調整
def fitIndexA(AB):
    for i in range(len(AB)):
        if 1 <= AB[i] and AB[i] <= 101:
            AB[i] -= 2
    return AB

#  fitIndexB
#    index範囲を調整
def fitIndexB(AB):
    for i in range(len(AB)):
        if -1 <= AB[i] and AB[i] <= 99:
            AB[i] += 1
    return AB

#
#  setDefaultGroupNo
#    defaultのgroupNoを予め設定し、preTet内からdefaultのgroupNoを削除する
#    defaultGroupは、group内の最大要素数のgroupNo。
def setDefaultGroupNo(elementList, preTet):
    defGrp = preTet[0][0]
    nElm = len(preTet[0][1])
    #要素数の多いgroupを取得
    for tet in preTet:
        if len(tet[1]) > nElm:
            nElm = len(tet[1])
            defGrp = tet[0]
    #要素数の多いgrpを削除
    newPreTet = []
    for tet in preTet:
        if defGrp != tet[0]:
            newPreTet.append(tet)
    #defaultのGrpNoを設定
    for i in range(len(elementList)):
        if len(elementList[i]) > 0:
            elementList[i][0][0] = defGrp
    return [elementList, newPreTet]

#
#  setElementGroups
#    elementListにgroupNoをセット
def setElementGroups(nodeList, elementList, preTet, minMaxLoc):
    #elementGroup数をチェック
    if len(preTet) == 1:
        #group数が1ヶの場合は、なにもせずに戻る
        return elementList
    #defaultのgroupNoを設定する
    [elementList, preTet] = setDefaultGroupNo(elementList, preTet)
    #要素のaddressを作成
    elmAdd = setElementAddress(nodeList, elementList, minMaxLoc)
    #設定or未設定用のflag
    elmsFlag = [0] * len(elementList)
    #preGroupのloop
    for tetra in preTet:
        grpNo = tetra[0]
        preData = tetra[1]
        #preのloop
        for preTet in preData:
            #blockを取得
            xx = [preTet[0][0], preTet[1][0], preTet[2][0], preTet[3][0]]
            yy = [preTet[0][1], preTet[1][1], preTet[2][1], preTet[3][1]]
            zz = [preTet[0][2], preTet[1][2], preTet[2][2], preTet[3][2]]
            minx = min(xx); maxx = max(xx)
            miny = min(yy); maxy = max(yy)
            minz = min(zz); maxz = max(zz)
            [ax, ay, az] = getLocAddress([minx, miny, minz], minMaxLoc)
            [bx, by, bz] = getLocAddress([maxx, maxy, maxz], minMaxLoc)
            #ax, bxを調整する
            [ax, ay, az] = fitIndexA([ax, ay, az])
            [bx, by, bz] = fitIndexB([bx, by, bz])
            [ax, ay, az] = fitIndexA([ax, ay, az])
            [bx, by, bz] = fitIndexB([bx, by, bz])
            #block内のAddressを取得
            blockElms = getLocAddInBlock(elmAdd, [ax, ay, az], [bx, by, bz])
            #block内のloop
            for elmNo in blockElms:
                #設定or未設定
                if elmsFlag[elmNo] == 0:
                    #要素の重心を取得
                    nodes = elementList[elmNo][0][1:5]
                    loc0 = nodeList[nodes[0]][0]
                    loc1 = nodeList[nodes[1]][0]
                    loc2 = nodeList[nodes[2]][0]
                    loc3 = nodeList[nodes[3]][0]
                    tetLoc = [loc0, loc1, loc2, loc3]
                    centerLoc = geo.tetG(tetLoc)
                    if geo.isPointInTetra(preTet, centerLoc) == True:
                        #grpNoとflagを設定
                        elementList[elmNo][0][0] = grpNo
                        elmsFlag[elmNo] = 1
    return elementList

#--------------------------
#  remakeSurfaceElementList
#--------------------------
#    surfaceList, elementListを修正する
def remakeSurfaceElementList(premeshFile, volFile):
    #preMeshの表面三角形、内部四面体の座標を取得
    [preTri, preTet, elSep, preNodeGrps] = readPreMeshConts(premeshFile)
    #volファイルを修正
    [nodeList, elementList, surfaceList] = getvol.readVolMesh(volFile)
    minMaxLoc = getMinMaxLocation(nodeList)
    #volファイル内のsurNo、elmNoを設定
    #  分割されたvolファイルは、surNo、elmNoが設定されていないので、
    #  premeshから読み取る
    surfaceList = setSurfaceGroups(nodeList, surfaceList, preTri, minMaxLoc)
    elementList = setElementGroups(nodeList, elementList, preTet, minMaxLoc)
    return [nodeList, elementList, surfaceList, minMaxLoc, elSep]

#
#  setGroupsFromPreMesh
#  --------------------
#    preMeshからgroupを取得し、volメッシュに設定
def setGroupsFromPreMesh(preMeshFile, volFileHeader):
    #preMeshの表面三角形、内部四面体の座標を取得
    [preTri, preTet] = readPreMeshConts(preMeshFile)
    #volファイルを修正
    volFile = volFileHeader + ".0.vol"
    [nodeList, elementList, surfaceList] = getvol.readVolMesh(volFile)
    surfaceList = setSurfaceGroups(nodeList, surfaceList, preTri)    
    elementList = setElementGroups(nodeList, elementList, preTet)


if __name__ == "__main__":
    import gettext
    gettext.install("app")

    preMeshFile = sys.argv[1]
    volFileHeader = sys.argv[2]
    setGroupsFromPreMesh(preMeshFile, volFileHeader)

